@@ -16,7 +16,7 @@ connects to the resource. One WorkflowController instance is connected to
 only one resource.
 
 The other classes (Workflow, Job, Group, FileTransfer and SharedResourcePath) are made to build the jobs, workflows, and file transfers
-objects to be used in the WorkflowControler interface.
+objects to be used in the WorkflowController interface.
 
 Definitions:
 A Parallel job is a job requiring more than one node to run.
@@ -35,13 +35,16 @@ import random
 import pickle
 import types
 import collections
+import subprocess
+import sys
 
 #import cProfile
+#import traceback
 
 import soma.workflow.connection as connection
 from soma.workflow.transfer import PortableRemoteTransfer, TransferSCP, TransferRsync, TransferMonitoring, TransferLocal
-from soma.workflow.constants import *
-from soma.workflow.configuration import Configuration
+import soma.workflow.constants as constants
+import soma.workflow.configuration as configuration
 from soma.workflow.errors import TransferError, SerializationError
 
 #-------------------------------------------------------------------------------
@@ -108,6 +111,10 @@ class Job(object):
     Path of the directory where the job will be executed. The working directory
     is useful if your Job uses relative file path for example.
 
+  **priority**: *int*
+    Job priority: 0 = low priority. If several Jobs are ready to run at the 
+    same time the jobs with higher priority will be submitted first.
+
   **parallel_job_info**: *tuple(string, int)*
     The parallel job information must be set if the Job is parallel (ie. made to
     run on several CPU).
@@ -157,6 +164,9 @@ class Job(object):
   # int (in hours)
   disposal_timeout = None
 
+  # int 
+  priority = None
+
   def __init__( self,
                 command,
                 referenced_input_files=None,
@@ -168,7 +178,8 @@ class Job(object):
                 stdout_file=None,
                 stderr_file=None,
                 working_directory=None,
-                parallel_job_info=None):
+                parallel_job_info=None,
+                priority=0):
     if not name:
       self.name = command[0]
     else:
@@ -187,6 +198,7 @@ class Job(object):
     self.stderr_file = stderr_file
     self.working_directory = working_directory
     self.parallel_job_info = parallel_job_info
+    self.priority = priority
 
 
 class Workflow(object):
@@ -232,15 +244,23 @@ class Workflow(object):
   # sequence of Groups built from the root_group
   groups = None
 
+  # for special user storage
+  user_storage = None
+
   def __init__(self,
                jobs,
-               dependencies,
-               root_group = None,
-               disposal_timeout = 168,):
+               dependencies=None,
+               root_group=None,
+               disposal_timeout=168,
+               user_storage=None,
+               name=None):
 
-    self.name = None
+    self.name = name
     self.jobs = jobs
-    self.dependencies = dependencies
+    if dependencies != None:
+      self.dependencies = dependencies
+    else:
+      self.dependencies = []
     self.disposal_timeout = 168
 
     # Groups
@@ -370,9 +390,9 @@ class FileTransfer(object):
     self.client_paths = client_paths
 
     if is_input:
-      self.initial_status = FILES_ON_CLIENT
+      self.initial_status = constants.FILES_ON_CLIENT
     else:
-      self.initial_status = FILES_DO_NOT_EXIST
+      self.initial_status = constants.FILES_DO_NOT_EXIST
 
 
 class SharedResourcePath(object):
@@ -433,21 +453,26 @@ class WorkflowController(object):
 
   config = None
 
+  engine_config_proxy = None
+
+  _resource_id = None
+
+  scheduler_config = None
+
   def __init__(self,
                resource_id=None,
                login=None,
                password=None,
-               config_path=None,
+               config=None,
                rsa_key_pass=None):
     '''
     Sets up the connection to the computing resource.
     Looks for a soma-workflow configuration file (if not specified in the
-    *config_path* argument).
+    *config* argument).
 
     * resource_id *string*
         Identifier of the computing resource to connect to.
-        The number of cpu of the current machine is detected and the basic
-        scheduler is lauched.
+        If None, the number of cpu of the current machine is detected and the basic scheduler is lauched.
 
     * login *string*
         Required if the computing resource is remote.
@@ -456,8 +481,8 @@ class WorkflowController(object):
         Required if the computing resource is remote and not RSA key where
         configured to log on the remote machine with ssh.
 
-    * config_path *string*
-        Optional path to the configuration file.
+    * config *configuration.Configuration*
+        Optional configuration.
 
     * rsa_key_pass *string*
         Required if the RSA key is protected with a password.
@@ -466,14 +491,25 @@ class WorkflowController(object):
       The login and password are only required for a remote computing resource.
     '''
 
-    self.config = Configuration(resource_id)
+    if config == None:
+      self.config = configuration.Configuration.load_from_file(resource_id)
+    else:
+      self.config = config
+
+    self.scheduler_config = None
+
     mode = self.config.get_mode()
     print  mode + " mode"
 
+    self._resource_id = resource_id
+    
+    
+
     # LOCAL MODE
-    if mode == 'local':
+    if mode == configuration.LOCAL_MODE:
       self._connection = connection.LocalConnection(resource_id, "")
       self._engine_proxy = self._connection.get_workflow_engine()
+      self.engine_config_proxy = self._connection.get_configuration()
       self._transfer = TransferLocal(self._engine_proxy)
       #self._transfer = PortableRemoteTransfer(self._engine_proxy)
       #self._transfer = TransferSCP(self._engine_proxy,
@@ -485,9 +521,10 @@ class WorkflowController(object):
                                     #hostname=None)
       self._transfer_stdouterr = TransferLocal(self._engine_proxy)
       #self._transfer = PortableRemoteTransfer(self._engine_proxy)
+      
 
     # REMOTE MODE
-    elif mode == 'remote':
+    elif mode == configuration.REMOTE_MODE:
       submitting_machines = self.config.get_submitting_machines()
       sub_machine = submitting_machines[random.randint(0,
                                                     len(submitting_machines)-1)]
@@ -502,6 +539,7 @@ class WorkflowController(object):
                                                     "",
                                                     rsa_key_pass)
       self._engine_proxy = self._connection.get_workflow_engine()
+      self.engine_config_proxy = self._connection.get_configuration()
 
       if not password and not rsa_key_pass:
         self._transfer = TransferSCP(self._engine_proxy,
@@ -512,8 +550,16 @@ class WorkflowController(object):
       self._transfer_stdouterr = PortableRemoteTransfer(self._engine_proxy)
 
     # LIGHT MODE
-    elif mode == 'light':
-      self._engine_proxy = _embedded_engine_and_server(self.config)
+    elif mode == configuration.LIGHT_MODE:
+      local_scdl_cfg_path = configuration.LocalSchedulerCfg.search_config_path()
+      if local_scdl_cfg_path == None:
+        cpu_count = Helper.cpu_count()
+        self.scheduler_config = configuration.LocalSchedulerCfg(proc_nb=cpu_count)
+      else:
+        self.scheduler_config = configuration.LocalSchedulerCfg.load_from_file(local_scdl_cfg_path)
+      self._engine_proxy = _embedded_engine_and_server(self.config, 
+                                                       self.scheduler_config)
+      self.engine_config_proxy = self.config
       self._connection = None
       self._transfer = TransferLocal(self._engine_proxy)
       #self._transfer = PortableRemoteTransfer(self._engine_proxy)
@@ -696,9 +742,9 @@ class WorkflowController(object):
 
     * workflow_id *workflow identifier*
 
-    * returns: tuple (sequence of tuple (job_id, status, exit_info,
+    * returns: tuple (sequence of tuple (job_id, status, queue, exit_info, 
       (submission_date, execution_date, ending_date)), sequence of tuple
-      (transfer_id, (status, progression_info)), workflow_status)
+      (transfer_id, (status, progression_info)), workflow_status, workflow_queue)
 
     Raises *UnknownObjectError* if the workflow_id is not valid
     '''
@@ -714,7 +760,7 @@ class WorkflowController(object):
 
       new_transfer_status.append((engine_path, (status, progression)))
 
-    new_wf_status = (wf_status[0], new_transfer_status, wf_status[2])
+    new_wf_status = (wf_status[0], new_transfer_status, wf_status[2], wf_status[3])
     return new_wf_status
 
 
@@ -825,7 +871,7 @@ class WorkflowController(object):
 
   ########## WORKFLOW CONTROL ############################################
 
-  def restart_workflow(self, workflow_id):
+  def restart_workflow(self, workflow_id, queue=None):
     '''
     Restarts the jobs of the workflow which failed. The jobs will be submitted
     again.
@@ -833,12 +879,16 @@ class WorkflowController(object):
 
     * workflow_id *workflow identifier*
 
+    * queue *string*
+        Optional name of the queue where to submit jobs. If it is not specified
+        the jobs will be submitted to the default queue.
+
     * returns: *boolean*
         True if some jobs were restarted.
 
     Raises *UnknownObjectError* if the workflow_id is not valid
     '''
-    return self._engine_proxy.restart_workflow(workflow_id)
+    return self._engine_proxy.restart_workflow(workflow_id, queue)
 
 
   def delete_workflow(self, workflow_id, force=True):
@@ -878,6 +928,12 @@ class WorkflowController(object):
     The running jobs will be killed.
     The jobs in queues will be removed from queues.
     It will be possible to restart the workflow afterwards.
+
+     * returns: *boolean*
+
+      return True if the running jobs were killed and False
+      if some jobs are possibly still running on the computing resource 
+      despite the workflow was stopped.
     '''
 
     return self._engine_proxy.stop_workflow(workflow_id)
@@ -915,7 +971,7 @@ class WorkflowController(object):
     '''
     self._engine_proxy.wait_job(job_ids, timeout)
 
-  def kill_job( self, job_id ):
+  def kill_job(self, job_id ):
     '''
     Kills a running job. The job will not be deleted from the system (the job
     identifier remains valid).
@@ -926,7 +982,7 @@ class WorkflowController(object):
     self._engine_proxy.kill_job(job_id)
 
 
-  def restart_job( self, job_id ):
+  def restart_job(self, job_id):
     '''
     Restarts a job which status is constants.FAILED or constants.WARNING.
 
@@ -939,16 +995,21 @@ class WorkflowController(object):
     self._engine_proxy.restart_job(job_id)
 
 
-  def delete_job( self, job_id):
+  def delete_job(self, job_id, force=True):
     '''
     Deletes a job which is not part of a workflow.
     The job_id will become invalid and can not be used anymore.
     The job is killed if it is running.
 
     Raises *UnknownObjectError* if the job_id is not valid
+
+    * returns: *boolean*
+      If force is True: return True if the running jobs were killed and False
+      if some jobs are possibly still running on the computing resource despite
+      the workflow doesn't exist.
     '''
 
-    self._engine_proxy.delete_job(job_id)
+    return self._engine_proxy.delete_job(job_id, force)
 
 
   ########## FILE TRANSFER CONTROL #######################################
@@ -1025,22 +1086,22 @@ class WorkflowController(object):
      transfer_type,
      status) = self._engine_proxy.transfer_information(transfer_id)
 
-    if status == FILES_ON_CLIENT:
+    if status == constants.FILES_ON_CLIENT:
       if not client_paths:
         if os.path.isfile(client_path):
-          transfer_type = TR_FILE_C_TO_CR
+          transfer_type = constants.TR_FILE_C_TO_CR
           self._engine_proxy.set_transfer_status(transfer_id,
-                                                 TRANSFERING_FROM_CLIENT_TO_CR)
+                                                 constants.TRANSFERING_FROM_CLIENT_TO_CR)
           self._engine_proxy.set_transfer_type(transfer_id,
                                                transfer_type)
 
 
         elif os.path.isdir(client_path):
-          transfer_type = TR_DIR_C_TO_CR
+          transfer_type = constants.TR_DIR_C_TO_CR
           self._engine_proxy.set_transfer_status(transfer_id,
-                                                 TRANSFERING_FROM_CLIENT_TO_CR)
+                                                 constants.TRANSFERING_FROM_CLIENT_TO_CR)
           self._engine_proxy.set_transfer_type(transfer_id,
-                                               TR_DIR_C_TO_CR)
+                                               constants.TR_DIR_C_TO_CR)
         else:
           print("WARNING: The file or directory %s doesn't exist "
                 "on the client machine." %(client_path))
@@ -1049,15 +1110,15 @@ class WorkflowController(object):
           if not os.path.isfile(path) and not os.path.isdir(path):
             print("WARNING: The file or directory %s doesn't exist "
                   "on the client machine." %(path))
-        transfer_type = TR_MFF_C_TO_CR
+        transfer_type = constants.TR_MFF_C_TO_CR
 
       self._engine_proxy.set_transfer_status(transfer_id,
-                                                TRANSFERING_FROM_CLIENT_TO_CR)
+                                             constants.TRANSFERING_FROM_CLIENT_TO_CR)
       self._engine_proxy.set_transfer_type(transfer_id,
                                            transfer_type)
       return transfer_type
 
-    elif status == FILES_ON_CR or status == FILES_ON_CLIENT_AND_CR:
+    elif status == constants.FILES_ON_CR or status == constants.FILES_ON_CLIENT_AND_CR:
       #transfer_type = self._engine_proxy.init_transfer_from_cr(transfer_id,
                                                #client_path,
                                                #expiration_date,
@@ -1066,9 +1127,9 @@ class WorkflowController(object):
                                                #status)
       if not client_paths:
         if self._engine_proxy.is_file(transfer_id):
-          transfer_type = TR_FILE_CR_TO_C
+          transfer_type = constants.TR_FILE_CR_TO_C
         elif self._engine_proxy.is_dir(transfer_id):
-          transfer_type = TR_DIR_CR_TO_C
+          transfer_type = constants.TR_DIR_CR_TO_C
         else:
           print("WARNING: The file or directory %s doesn't exist "
                 "on the computing resource side." %(transfer_id))
@@ -1080,10 +1141,10 @@ class WorkflowController(object):
              not self._engine_proxy.is_dir(r_path):
             print("WARNING: The file or directory %s doesn't exist "
                   "on the computing resource side." %(r_path))
-        transfer_type = TR_MFF_CR_TO_C
+        transfer_type = constants.TR_MFF_CR_TO_C
 
       self._engine_proxy.set_transfer_status(transfer_id,
-                                        TRANSFERING_FROM_CR_TO_CLIENT)
+                                        constants.TRANSFERING_FROM_CR_TO_CLIENT)
       self._engine_proxy.set_transfer_type(transfer_id,
                                            transfer_type)
 
@@ -1100,30 +1161,30 @@ class WorkflowController(object):
      transfer_type,
      status) = self._engine_proxy.transfer_information(transfer_id)
 
-    if status == FILES_ON_CLIENT or \
-       status == TRANSFERING_FROM_CLIENT_TO_CR:
+    if status == constants.FILES_ON_CLIENT or \
+       status == constants.TRANSFERING_FROM_CLIENT_TO_CR:
       # transfer from client to computing resource
       #overwrite = False
       #if not transfer_type or \
-         #transfer_type == TR_FILE_CR_TO_C or \
-         #transfer_type == TR_DIR_CR_TO_C or \
-         #transfer_type == TR_MFF_CR_TO_C:
+         #transfer_type == constants.TR_FILE_CR_TO_C or \
+         #transfer_type == constants.TR_DIR_CR_TO_C or \
+         #transfer_type == constants.TR_MFF_CR_TO_C:
         ## transfer reset
         #overwrite = True
       transfer_type = self._initialize_transfer(transfer_id)
 
       remote_path = transfer_id
 
-      if transfer_type == TR_FILE_C_TO_CR or \
-         transfer_type == TR_DIR_C_TO_CR:
+      if transfer_type == constants.TR_FILE_C_TO_CR or \
+         transfer_type == constants.TR_DIR_C_TO_CR:
         self._transfer.transfer_to_remote(client_path,
                                          remote_path)
         self._engine_proxy.set_transfer_status(transfer_id,
-                                               FILES_ON_CLIENT_AND_CR)
+                                               constants.FILES_ON_CLIENT_AND_CR)
         self._engine_proxy.signalTransferEnded(transfer_id, workflow_id)
         return True
 
-      if transfer_type == TR_MFF_C_TO_CR:
+      if transfer_type == constants.TR_MFF_C_TO_CR:
         for path in client_paths:
           relative_path = os.path.basename(path)
           r_path = os.path.join(remote_path, relative_path)
@@ -1131,36 +1192,36 @@ class WorkflowController(object):
                                             r_path)
 
         self._engine_proxy.set_transfer_status(transfer_id,
-                                               FILES_ON_CLIENT_AND_CR)
+                                               constants.FILES_ON_CLIENT_AND_CR)
         self._engine_proxy.signalTransferEnded(transfer_id, workflow_id)
         return True
 
-    if status == FILES_ON_CR or \
-       status == TRANSFERING_FROM_CR_TO_CLIENT or \
-       status == FILES_ON_CLIENT_AND_CR:
+    if status == constants.FILES_ON_CR or \
+       status == constants.TRANSFERING_FROM_CR_TO_CLIENT or \
+       status == constants.FILES_ON_CLIENT_AND_CR:
       # transfer from computing resource to client
       #overwrite = False
       #if not transfer_type or \
-         #transfer_type == TR_FILE_C_TO_CR or \
-         #transfer_type == TR_DIR_C_TO_CR or \
-         #transfer_type == TR_MFF_C_TO_CR :
+         #transfer_type == constants.TR_FILE_C_TO_CR or \
+         #transfer_type == constants.TR_DIR_C_TO_CR or \
+         #transfer_type == constants.TR_MFF_C_TO_CR :
         ## TBI remove existing files
         #overwrite = True
       transfer_type = self._initialize_transfer(transfer_id)
 
 
       remote_path = transfer_id
-      if transfer_type == TR_FILE_CR_TO_C or \
-         transfer_type == TR_DIR_CR_TO_C:
+      if transfer_type == constants.TR_FILE_CR_TO_C or \
+         transfer_type == constants.TR_DIR_CR_TO_C:
         # file case
         self._transfer.transfer_from_remote(remote_path,
                                             client_path)
         self._engine_proxy.set_transfer_status(transfer_id,
-                                               FILES_ON_CLIENT_AND_CR)
+                                               constants.FILES_ON_CLIENT_AND_CR)
         self._engine_proxy.signalTransferEnded(transfer_id, workflow_id)
         return True
 
-      if transfer_type == TR_MFF_CR_TO_C:
+      if transfer_type == constants.TR_MFF_CR_TO_C:
         for path in client_paths:
           relative_path = os.path.basename(path)
           r_path = os.path.join(remote_path, relative_path)
@@ -1168,7 +1229,7 @@ class WorkflowController(object):
                                               path)
 
         self._engine_proxy.set_transfer_status(transfer_id,
-                                               FILES_ON_CLIENT_AND_CR)
+                                               constants.FILES_ON_CLIENT_AND_CR)
         self._engine_proxy.signalTransferEnded(transfer_id, workflow_id)
         return True
 
@@ -1181,8 +1242,8 @@ class WorkflowController(object):
                             client_path,
                             client_paths,
                             engine_path):
-    if status == TRANSFERING_FROM_CLIENT_TO_CR:
-      if transfer_type == TR_MFF_C_TO_CR:
+    if status == constants.TRANSFERING_FROM_CLIENT_TO_CR:
+      if transfer_type == constants.TR_MFF_C_TO_CR:
         data_size = 0
         data_transfered = 0
         for path in client_paths:
@@ -1199,8 +1260,8 @@ class WorkflowController(object):
                                                                 client_path,
                                                                 engine_path)
 
-    elif status == TRANSFERING_FROM_CR_TO_CLIENT:
-      if transfer_type == TR_MFF_CR_TO_C:
+    elif status == constants.TRANSFERING_FROM_CR_TO_CLIENT:
+      if transfer_type == constants.TR_MFF_CR_TO_C:
         data_size = 0
         data_transfered = 0
         for path in client_paths:
@@ -1221,29 +1282,7 @@ class WorkflowController(object):
     return progression
 
 
-def _detectCPUs():
- """
- Detects the number of CPUs on a system.
- ==> Python >= 2.6: multiprocessing.cpu_count
- """
- # Linux, Unix and MacOS:
- if hasattr(os, "sysconf"):
-     if os.sysconf_names.has_key("SC_NPROCESSORS_ONLN"):
-         # Linux & Unix:
-         ncpus = os.sysconf("SC_NPROCESSORS_ONLN")
-         if isinstance(ncpus, int) and ncpus > 0:
-             return ncpus
-     else: # OSX:
-         return int(os.popen2("sysctl -n hw.ncpu")[1].read())
- # Windows:
- if os.environ.has_key("NUMBER_OF_PROCESSORS"):
-         ncpus = int(os.environ["NUMBER_OF_PROCESSORS"]);
-         if ncpus > 0:
-             return ncpus
- return 1 # Default
-
-
-def _embedded_engine_and_server(config):
+def _embedded_engine_and_server(config, local_scheduler_config=None):
   '''
   Creates the workflow engine and workflow database server in the client
   process.
@@ -1257,7 +1296,7 @@ def _embedded_engine_and_server(config):
   '''
   import logging
 
-  from soma.workflow.engine import WorkflowEngine, WorkflowEngineLoop, EngineLoopThread
+  from soma.workflow.engine import WorkflowEngine, ConfiguredWorkflowEngine
   from soma.workflow.database_server import WorkflowDatabaseServer
 
   (engine_log_dir,
@@ -1285,22 +1324,15 @@ def _embedded_engine_and_server(config):
                       config.get_parallel_job_config())
 
   elif config.get_scheduler_type() == 'local_basic':
-    from soma.workflow.scheduler import LocalScheduler
-    cpu_count = _detectCPUs()
-    print "scheduler type: basic, cpu count: " + repr(cpu_count)
-    scheduler = LocalScheduler(nb_proc=cpu_count)
+    from soma.workflow.scheduler import ConfiguredLocalScheduler
+    if local_scheduler_config == None:
+      local_scheduler_config = LocalSchedulerCfg()
+    print "scheduler type: basic, number of cpu: " + repr(local_scheduler_config.get_proc_nb())
+    scheduler = ConfiguredLocalScheduler(local_scheduler_config)
 
-  engine_loop = WorkflowEngineLoop(database_server,
-                                   scheduler,
-                                   config.get_path_translation(),
-                                   config.get_queue_limits())
-
-  workflow_engine = WorkflowEngine(database_server,
-                                   engine_loop)
-
-  engine_loop_thread = EngineLoopThread(engine_loop)
-  engine_loop_thread.setDaemon(True)
-  engine_loop_thread.start()
+  workflow_engine = ConfiguredWorkflowEngine(database_server,
+                                             scheduler, 
+                                             config)
 
   return workflow_engine
 
@@ -1333,13 +1365,13 @@ class Helper(object):
 
 
   @staticmethod
-  def transfer_input_files(workflow,
+  def transfer_input_files(workflow_id,
                            wf_ctrl,
                            buffer_size = 512**2):
     '''
     Transfers all the input files of a workflow.
 
-    * workflow *client.Workflow*
+    * workflow_id *workflow identifier*
 
     * wf_ctrl *client.WorkflowController*
 
@@ -1348,58 +1380,60 @@ class Helper(object):
         piece. The size of each piece can be tuned using the buffer_size
         argument.
     '''
-
     transfer_info = None
-    wf_elements_status = wf_ctrl.workflow_elements_status(workflow.wf_id)
+    wf_elements_status = wf_ctrl.workflow_elements_status(workflow_id)
 
     to_transfer = []
     for transfer_info in wf_elements_status[1]:
       status = transfer_info[1][0]
-      if status == FILES_ON_CLIENT:
+      if status == constants.FILES_ON_CLIENT:
         engine_path = transfer_info[0]
         to_transfer.append(engine_path)
-      if status == TRANSFERING_FROM_CLIENT_TO_CR:
+      if status == constants.TRANSFERING_FROM_CLIENT_TO_CR:
         engine_path = transfer_info[0]
         to_transfer.append(engine_path)
 
-    for engine_path in to_transfer:
-      wf_ctrl.transfer_files(engine_path, buffer_size)
+    wf_ctrl.transfer_files(to_transfer, buffer_size)
 
 
 
   @staticmethod
-  def transfer_output_files(workflow,
+  def transfer_output_files(workflow_id,
                             wf_ctrl,
-                            buffer_size = 512**2):
+                            buffer_size=512**2):
     '''
     Transfers all the output files of a workflow which are ready to transfer.
 
-    * workflow *client.Workflow*
+    * workflow_id *workflow identifier*
 
     * wf_ctrl *client.WorkflowController*
-    '''
 
+    * buffer_size *int*
+        Depending on the transfer method, the files can be transfered piece by
+        piece. The size of each piece can be tuned using the buffer_size
+        argument.
+    '''
     transfer_info = None
-    wf_elements_status = wf_ctrl.workflow_elements_status(workflow.wf_id)
+    wf_elements_status = wf_ctrl.workflow_elements_status(workflow_id)
 
     to_transfer = []
     for transfer_info in wf_elements_status[1]:
       status = transfer_info[1][0]
-      if status == FILES_ON_CR:
+      if status == constants.FILES_ON_CR:
         engine_path = transfer_info[0]
         to_transfer.append(engine_path)
-      if status == TRANSFERING_FROM_CR_TO_CLIENT:
+      if status == constants.TRANSFERING_FROM_CR_TO_CLIENT:
         engine_path = transfer_info[0]
         to_transfer.append(engine_path)
 
-    for engine_path in to_transfer:
-      wf_ctrl.transfer_files(engine_path, buffer_size)
+    wf_ctrl.transfer_files(to_transfer, buffer_size)
+
 
 
   @staticmethod
   def serialize(file_path, workflow):
     '''
-    Saves a workflow to a file
+    Saves a workflow to a file => Python Pickle for now.
 
     * file_path *String*
 
@@ -1417,7 +1451,7 @@ class Helper(object):
   @staticmethod
   def unserialize(file_path):
     '''
-    Loads a workflow from a file
+    Loads a workflow from a file 
 
     * file_path *String*
 
@@ -1434,5 +1468,34 @@ class Helper(object):
     return workflow
 
 
-
+  @staticmethod
+  def cpu_count():
+    """
+    Detects the number of CPUs on a system.
+    ==> Python >= 2.6: multiprocessing.cpu_count
+    """
+    if sys.version_info[:2] >= (2,6):
+      try:
+        import multiprocessing
+        return multiprocessing.cpu_count()
+      except: # sometimes happens on MacOS... ?
+        print >> sys.stderr, \
+            'Warning: CPU count detection failed. Using default (2)'
+        return 2
+    # Linux, Unix and MacOS:
+    if hasattr(os, "sysconf"):
+        if os.sysconf_names.has_key("SC_NPROCESSORS_ONLN"):
+            # Linux & Unix:
+            ncpus = os.sysconf("SC_NPROCESSORS_ONLN")
+            if isinstance(ncpus, int) and ncpus > 0:
+                return ncpus
+        else: # OSX:
+            return int(subprocess.Popen(["sysctl", "-n", "hw.ncpu"],
+                stdout=subprocess.PIPE).stdout.read())
+    # Windows:
+    if os.environ.has_key("NUMBER_OF_PROCESSORS"):
+            ncpus = int(os.environ["NUMBER_OF_PROCESSORS"]);
+            if ncpus > 0:
+                return ncpus
+    return 1 # Default
 
