@@ -23,15 +23,17 @@ import time
 import logging
 import stat, hashlib, operator
 import itertools
-
+import atexit
 
 #import cProfile
+#import traceback
 
 from soma.workflow.engine_types import EngineJob, EngineWorkflow, EngineTransfer
 import soma.workflow.constants as constants
 from soma.workflow.client import WorkflowController
 from soma.workflow.errors import JobError, UnknownObjectError, EngineError, DRMError
 from soma.workflow.transfer import RemoteFileController
+from soma.workflow.configuration import Configuration
 
 #-----------------------------------------------------------------------------
 # Globals and constants
@@ -43,6 +45,20 @@ refreshment_interval = 1 #seconds
 # the status is changed into WARNING
 refreshment_timeout = 60 #seconds
 
+def _out_to_date(last_status_update):
+  '''
+  Tells if a workflow or a job is out to date considering its last  
+  status update. 
+
+  * last_status_update: *datetime*
+  '''
+  if last_status_update is None:
+    return False
+  delta = datetime.now() - last_status_update
+  out_to_date = delta > timedelta(seconds=refreshment_timeout)
+  return out_to_date
+
+
 #-----------------------------------------------------------------------------
 # Classes and functions
 #-----------------------------------------------------------------------------
@@ -52,6 +68,7 @@ class EngineLoopThread(threading.Thread):
     super(EngineLoopThread, self).__init__()
     self.engine_loop = engine_loop
     self.time_interval = refreshment_interval
+    atexit.register(EngineLoopThread.stop, self)
   
   def run(self):
     #cProfile.runctx("self.engine_loop.start_loop(self.time_interval)", globals(), locals(), "/home/sl225510/profiling/profile_loop_thread")
@@ -60,9 +77,9 @@ class EngineLoopThread(threading.Thread):
   def stop(self):
     self.engine_loop.stop_loop()
     self.join()
+    print "Soma workflow engine thread ended nicely."
   
 
-
 class WorkflowEngineLoop(object):
 
   # jobs managed by the current engine process instance. 
@@ -151,6 +168,7 @@ class WorkflowEngineLoop(object):
     #one_wf_processed = False
     self._running = True
     drms_error_jobs = {}
+    idle_cmpt = 0
     while True:
       if not self._running:
         break
@@ -161,6 +179,15 @@ class WorkflowEngineLoop(object):
           if job.workflow_id != -1: 
             wf_to_inspect.add(job.workflow_id)
         
+        if not (len(self._jobs) == 0 and len(self._workflows) == 0):
+          idle_cmpt = 0
+        else:
+          idle_cmpt = idle_cmpt + 1
+
+        if idle_cmpt > 20 and not self._scheduler.is_sleeping:
+          self.logger.debug("idle => scheduler sleep")
+          self._scheduler.sleep()
+
 
         # --- 1. Jobs and workflow deletion and kill ------------------------
         # Get the jobs and workflow with the status DELETE_PENDING 
@@ -182,7 +209,7 @@ class WorkflowEngineLoop(object):
               stopped = self._stop_job(job_id,  self._jobs[job_id])
             except DRMError, e:
               #TBI how to communicate the error ?
-              self.logger.error("!!!ERROR!!! %s :%s" %(type(e), e))
+              self.logger.error("!!!ERROR!!! stop job %s :%s" %(type(e), e))
             if job_id in jobs_to_delete:
               self.logger.debug("Delete job : " + repr(job_id))
               self._database_server.delete_job(job_id)
@@ -224,7 +251,7 @@ class WorkflowEngineLoop(object):
             try:
               job.status = self._scheduler.get_job_status(job.drmaa_id)
             except DrmaaError, e:
-              self.logger.error("!!!ERROR!!! %s: %s" %(type(e), e))
+              self.logger.error("!!!ERROR!!! get_job_status %s: %s" %(type(e), e))
               job.status = constants.UNDETERMINED
             #self.logger.debug("job " + repr(job.job_id) + " : " + job.status)
             if job.status == constants.DONE or job.status == constants.FAILED:
@@ -340,6 +367,10 @@ class WorkflowEngineLoop(object):
   def stop_loop(self):
     self._running = False
 
+  def set_queue_limits(self, queue_limits):
+    with self._lock:
+      self._queue_limits = queue_limits
+
   def add_job(self, client_job, queue):
     # register
     engine_job = EngineJob(client_job=client_job, 
@@ -393,6 +424,8 @@ class WorkflowEngineLoop(object):
     '''
     if engine_job.queue in self._pending_queues:
       self._pending_queues[engine_job.queue].append(engine_job)
+      self._pending_queues[engine_job.queue].sort(key=lambda job: job.priority, 
+                                                  reverse=True)
     else:
       self._pending_queues[engine_job.queue] = [engine_job]
     engine_job.status = constants.SUBMISSION_PENDING
@@ -513,23 +546,36 @@ class WorkflowEngineLoop(object):
     return ended_jobs
 
 
-  def restart_workflow(self, wf_id, status):
+  def restart_workflow(self, wf_id, status, queue):
     if wf_id in self._workflows:
       workflow = self._workflows[wf_id]
+      workflow.queue = queue
       (jobs_to_run, 
-       workflow.status) = workflow.restart(self._database_server)
+       workflow.status) = workflow.restart(self._database_server, queue)
       for job in jobs_to_run:
         self._pend_for_submission(job)
     else:
       workflow = self._database_server.get_engine_workflow(wf_id, self._user_id)
       workflow.status = status
-      (jobs_to_run, workflow.status) = workflow.restart(self._database_server)
+      (jobs_to_run, workflow.status) = workflow.restart(self._database_server, queue)
       for job in jobs_to_run:
         self._pend_for_submission(job)
       # add to the engine managed workflow list
       with self._lock:
         self._workflows[wf_id] = workflow
 
+  def force_stop(self, wf_id):
+    if wf_id in self._workflows:
+      pass
+    else:
+      workflow = self._database_server.get_engine_workflow(wf_id,
+                                                           self._user_id)
+      workflow.force_stop(self._database_server)
+      with self._lock:
+        self._workflows[wf_id] = workflow
+      
+
+
   def restart_job(self, job_id, status):
     (job, workflow_id) = self._database_server.get_engine_job(job_id, self._user_id)
     if workflow_id == -1: 
@@ -552,12 +598,17 @@ class WorkflowEngine(RemoteFileController):
   # soma.workflow.database_server.WorkflowDatabaseServer
   _database_server = None
   # WorkflowEngineLoop
-  _engine_loop = None
+  engine_loop = None
+  # EngineLoopThread
+  engine_loop_thread = None
   # id of the user on the database server
   _user_id = None
 
-
-  def __init__( self, database_server, engine_loop):
+  def __init__( self, 
+                database_server, 
+                scheduler, 
+                path_translation=None,
+                queue_limits={}):
     ''' 
     @type  database_server:
            L{soma.workflow.database_server.WorkflowDatabaseServer}
@@ -567,7 +618,7 @@ class WorkflowEngine(RemoteFileController):
     self.logger = logging.getLogger('engine.WorkflowEngine')
     
     self._database_server= database_server
-    self._engine_loop = engine_loop
+    
     
     try:
       user_login = pwd.getpwuid(os.getuid())[0]
@@ -577,6 +628,14 @@ class WorkflowEngine(RemoteFileController):
     self._user_id = self._database_server.register_user(user_login)
     self.logger.debug("user_id : " + repr(self._user_id))
 
+    self.engine_loop = WorkflowEngineLoop(database_server,
+                                           scheduler,
+                                           path_translation,
+                                           queue_limits)
+    self.engine_loop_thread = EngineLoopThread(self.engine_loop)
+    self.engine_loop_thread.setDaemon(True)
+    self.engine_loop_thread.start()
+
   def __del__( self ):
     pass
 
@@ -684,12 +743,12 @@ class WorkflowEngine(RemoteFileController):
     @type  job: L{soma.workflow.client.Job}
     @param job: job informations 
     '''
-    engine_job = self._engine_loop.add_job(job, queue)
+    engine_job = self.engine_loop.add_job(job, queue)
 
     return engine_job.job_id
 
 
-  def delete_job( self, job_id ):
+  def delete_job( self, job_id, force=True ):
     '''
     Implementation of soma.workflow.client.WorkflowController API
     '''
@@ -697,12 +756,14 @@ class WorkflowEngine(RemoteFileController):
                                                   self._user_id)[0]
     if status == constants.DONE or status == constants.FAILED:
       self._database_server.delete_job(job_id)
+      return True
     else:
       self._database_server.set_job_status(job_id, constants.DELETE_PENDING)
-      if not self._wait_for_job_deletion(job_id):
-        # TBI
+      if force and not self._wait_for_job_deletion(job_id):
         self.logger.critical("!! The job may not be properly deleted !!")
         self._database_server.delete_job(job_id)
+        return False
+      return True
 
   ########## WORKFLOW SUBMISSION ############################################
   
@@ -713,7 +774,7 @@ class WorkflowEngine(RemoteFileController):
     if not expiration_date:
       expiration_date = datetime.now() + timedelta(days=7)
     
-    wf_id = self._engine_loop.add_workflow(workflow, expiration_date, name, queue)
+    wf_id = self.engine_loop.add_workflow(workflow, expiration_date, name, queue)
 
     return wf_id
 
@@ -737,13 +798,23 @@ class WorkflowEngine(RemoteFileController):
        return False
       return True
 
+
   def stop_workflow(self, workflow_id):
 
-    status = self._database_server.get_workflow_status(workflow_id, 
-                                                       self._user_id)[0]
+    (status,  
+     last_status_update) = self._database_server.get_workflow_status(workflow_id, self._user_id)
+
     if status != constants.WORKFLOW_DONE:
-      self._database_server.set_workflow_status(workflow_id, 
-                                                constants.KILL_PENDING)
+      if _out_to_date(last_status_update):
+        self.engine_loop.force_stop(workflow_id)
+        return False
+      else:
+        self._database_server.set_workflow_status(workflow_id, 
+                                                  constants.KILL_PENDING)
+        self._wait_wf_status_update(workflow_id, 
+                                    expected_status = constants.WORKFLOW_DONE)
+
+    return True
 
 
   def change_workflow_expiration_date(self, workflow_id, new_expiration_date):
@@ -760,17 +831,19 @@ class WorkflowEngine(RemoteFileController):
     return True
 
 
-  def restart_workflow(self, workflow_id):
+  def restart_workflow(self, workflow_id, queue):
     '''
     Implementation of soma.workflow.client.WorkflowController API
     '''
     (status, last_status_update) = self._database_server.get_workflow_status(workflow_id, self._user_id)
     
-    if status != constants.WORKFLOW_DONE and status != constants.WARNING:
+    if status == constants.WORKFLOW_DONE:
+      self.engine_loop.restart_workflow(workflow_id, status, queue)
+      self._wait_wf_status_update(workflow_id, 
+                                  expected_status=constants.WORKFLOW_IN_PROGRESS)
+      return True
+    else:
       return False
-
-    self._engine_loop.restart_workflow(workflow_id, status)
-    return True
     
    
   ########## SERVER STATE MONITORING ########################################
@@ -813,9 +886,7 @@ class WorkflowEngine(RemoteFileController):
                                                                self._user_id)
     if status and not status == constants.DONE and \
        not status == constants.FAILED and \
-       last_status_update and \
-       datetime.now() - last_status_update > timedelta(seconds = refreshment_interval*refreshment_timeout):
-      self._database_server.set_job_status(job_id, constants.WARNING)
+       _out_to_date(last_status_update):
       return constants.WARNING
 
     return status
@@ -827,13 +898,11 @@ class WorkflowEngine(RemoteFileController):
     '''
     (status, 
      last_status_update) = self._database_server.get_workflow_status(wf_id,                                                              
-                                                                 self._user_id)
+                                                               self._user_id)
 
     if status and \
        not status == constants.WORKFLOW_DONE and \
-       last_status_update and \
-       datetime.now() - last_status_update > timedelta(seconds = refreshment_interval*refreshment_timeout):
-      self._database_server.set_workflow_status(wf_id, constants.WARNING)
+       _out_to_date(last_status_update):
       return constants.WARNING
 
     return status
@@ -846,13 +915,13 @@ class WorkflowEngine(RemoteFileController):
     (status, 
      last_status_update) = self._database_server.get_workflow_status(wf_id,
                                                                   self._user_id)
+    
+    wf_status = self._database_server.get_detailed_workflow_status(wf_id)
     if status and \
        not status == constants.WORKFLOW_DONE and \
-       last_status_update and \
-       datetime.now() - last_status_update > timedelta(seconds = refreshment_interval*refreshment_timeout):
-      self._database_server.set_workflow_status(wf_id, constants.WARNING)
-
-    wf_status = self._database_server.get_detailed_workflow_status(wf_id)
+       _out_to_date(last_status_update):
+      wf_status = (wf_status[0], wf_status[1], constants.WARNING, wf_status[3])
+    
     return wf_status
         
         
@@ -898,7 +967,6 @@ class WorkflowEngine(RemoteFileController):
       if status:
         self.logger.debug("wait        job %s status: %s", jid, status)
         delta = datetime.now()-startTime
-        delta_status_update = datetime.now() - last_status_update
         while status and not status == constants.DONE and not status == constants.FAILED and (waitForever or delta < timedelta(seconds=timeout)):
           time.sleep(refreshment_interval)
           (status, last_status_update) = self._database_server.get_job_status(jid, self._user_id) 
@@ -909,7 +977,7 @@ class WorkflowEngine(RemoteFileController):
                             repr(last_status_update), 
                             repr(datetime.now()))
           delta = datetime.now() - startTime
-          if last_status_update and datetime.now() - last_status_update > timedelta(seconds = refreshment_interval*refreshment_timeout):
+          if last_status_update and _out_to_date(last_status_update):
             raise EngineError("wait_job: Could not wait for job %s. " 
                               "The process updating its status failed." %(jid))
        
@@ -923,113 +991,143 @@ class WorkflowEngine(RemoteFileController):
      last_status_update) = self._database_server.get_job_status(job_id,
                                                                 self._user_id)
     
-    if status != constants.FAILED and status != constants.WARNING:
+    if status == constants.FAILED or _out_to_date(last_status_update):
+      self.engine_loop.restart_job(job_id, status)
+      return True
+    else:
       return False
 
-    self._engine_loop.restart_job(job_id, status)
-    return True
-
 
   def kill_job( self, job_id ):
     '''
     Implementation of soma.workflow.client.WorkflowController API
     '''
     
-    status = self._database_server.get_job_status(job_id,
-                                                  self._user_id)[0]
+    (status,
+    last_status_update) = self._database_server.get_job_status(job_id,
+                                                  self._user_id)
+
     if status != constants.DONE and status != constants.FAILED:
-      self._database_server.set_job_status(job_id, 
-                                          constants.KILL_PENDING)
-      
-      if not self._wait_job_status_update(job_id):
+      if _out_to_date(last_status_update):
+        # TO DO
+        pass
+      else:
         self._database_server.set_job_status(job_id, 
-                                             constants.WARNING)
+                                             constants.KILL_PENDING)
+      
+      self._wait_job_status_update(job_id)
 
 
-  def _wait_for_job_deletion(self, job_id):
-    self.logger.debug(">> _wait_for_job_deletion")
-    action_time = datetime.now()
-    time.sleep(refreshment_interval)
-    (is_valid_job, 
-     last_status_update) = self._database_server.is_valid_job(job_id, 
-                                                              self._user_id)
-    while is_valid_job and \
-          last_status_update < action_time:
-      time.sleep(refreshment_interval)
-      (is_valid_job, 
-       last_status_update) = self._database_server.is_valid_job(job_id, 
-                                                                self._user_id)
-      if last_status_update and datetime.now() - last_status_update > timedelta(seconds = refreshment_interval*refreshment_timeout):
-        self.logger.debug("<< _wait_for_job_deletion")
-        return False
-
-    self.logger.debug("<< _wait_for_job_deletion")
-    return True
-
 
   def _wait_job_status_update(self, job_id):
-    
     self.logger.debug(">> _wait_job_status_update")
-    action_time = datetime.now()
-    time.sleep(refreshment_interval)
     try:
       (status, 
       last_status_update) = self._database_server.get_job_status(job_id,
-                                                                  self._user_id)
+                                                                 self._user_id)
       while status and not status == constants.DONE and \
             not status == constants.FAILED and \
-            last_status_update < action_time:
+            not _out_to_date(last_status_update):
         time.sleep(refreshment_interval)
         (status, 
         last_status_update) = self._database_server.get_job_status(job_id,
                                                                   self._user_id) 
-        if last_status_update and datetime.now() - last_status_update > timedelta(seconds = refreshment_interval*refreshment_timeout):
-          self.logger.debug("<< _wait_job_status_update")
-          return False
     except UnknownObjectError, e:
       pass
     self.logger.debug("<< _wait_job_status_update")
+
+
+  def _wait_for_job_deletion(self, job_id):
+    self.logger.debug(">> _wait_for_job_deletion")
+    (is_valid_job, 
+     last_status_update) = self._database_server.is_valid_job(job_id, 
+                                                              self._user_id)
+    while is_valid_job and not _out_to_date(last_status_update):
+      time.sleep(refreshment_interval)
+      (is_valid_job, 
+       last_status_update) = self._database_server.is_valid_job(job_id, 
+                                                                self._user_id)
+    self.logger.debug("<< _wait_for_job_deletion")
+    if _out_to_date(last_status_update):
+      return False
     return True
 
-  def _wait_wf_status_update(self, wf_id):  
+
+  def _wait_wf_status_update(self, wf_id, expected_status):  
     self.logger.debug(">> _wait_wf_status_update")
-    action_time = datetime.now()
-    time.sleep(refreshment_interval)
     try:
       (status, 
-      last_status_update) = self._database_server.get_workflow_status(wf_id)
-      while status and not status == constants.WORKFLOW_DONE and \
-            last_status_update < action_time:
+      last_status_update) = self._database_server.get_workflow_status(wf_id,
+                                                                 self._user_id)
+      if status != None and status != expected_status:
         time.sleep(refreshment_interval)
         (status, 
-        last_status_update) = self._database_server.get_workflow_status(wf_id) 
-        if last_status_update and \
-          datetime.now() - last_status_update > timedelta(seconds=refreshment_interval*refreshment_timeout):
-          self.logger.debug("<< _wait_wf_status_update")
-          return False
+        last_status_update) = self._database_server.get_workflow_status(wf_id,
+                                                                  self._user_id) 
+        if status != None and status != expected_status:
+          time.sleep(refreshment_interval)
+          (status, 
+          last_status_update) = self._database_server.get_workflow_status(wf_id,
+                                                                    self._user_id)
+          while status != None and \
+                status != expected_status and \
+                status != constants.WORKFLOW_DONE and \
+                not _out_to_date(last_status_update):
+            time.sleep(refreshment_interval)
+            (status, 
+            last_status_update) = self._database_server.get_workflow_status(wf_id,
+                                                                      self._user_id) 
     except UnknownObjectError, e:
       pass
     self.logger.debug("<< _wait_wf_status_update")
-    return True
 
 
   def _wait_for_wf_deletion(self, wf_id):
     self.logger.debug(">> _wait_for_wf_deletion")
-    action_time = datetime.now()
-    time.sleep(refreshment_interval)
     (is_valid_wf, 
     last_status_update) = self._database_server.is_valid_workflow(wf_id, 
                                                                 self._user_id)
     while is_valid_wf and \
-          last_status_update < action_time:
+          not _out_to_date(last_status_update):
       time.sleep(refreshment_interval)
       (is_valid_wf, 
       last_status_update) = self._database_server.is_valid_workflow(wf_id, 
-                                                              self._user_id)
-      if last_status_update and datetime.now() - last_status_update > timedelta(seconds = refreshment_interval*refreshment_timeout):
-        self.logger.debug("<< _wait_for_wf_deletion")
-        return False
+                                                              self._user_id)     
     self.logger.debug("<< _wait_for_wf_deletion")
+    if _out_to_date(last_status_update):
+      return False
     return True
+
+class ConfiguredWorkflowEngine(WorkflowEngine):
+  '''
+  WorkflowEngineLoop synchronized with a configuration object.
+  '''
+
+  config = None
+  
+  def __init__(self, 
+               database_server, 
+               scheduler, 
+               config):
+    '''
+    * config *configuration.Configuration*
+    '''
+    super(ConfiguredWorkflowEngine, self).__init__(
+                               database_server,
+                               scheduler,
+                               path_translation=config.get_path_translation(),
+                               queue_limits=config.get_queue_limits())
+
+    self.config = config
+    
+    self.config.addObserver(self,
+                            "update_from_config",
+                            [Configuration.QUEUE_LIMITS_CHANGED])
+
+  def update_from_config(self, observable, event, msg):
+    if event == Configuration.QUEUE_LIMITS_CHANGED:
+      self.engine_loop.set_queue_limits(self.config.get_queue_limits())
+    self.config.save_to_file()
+
     
     
