@@ -1,12 +1,25 @@
 from __future__ import with_statement
 
+'''
+@author: Soizic Laguitton
+@organization: U{IFR 49<http://www.ifr49.org>}
+@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
+'''
+
 import subprocess
 import threading
 import time
 import logging
 import os
+import sys
+import signal
+import ctypes
+import atexit
+import os.path
 
 import soma.workflow.constants as constants
+from soma.workflow.errors import DRMError
+from soma.workflow.configuration import LocalSchedulerCfg
 
 try:
   from soma.workflow.somadrmaajobssip import DrmaaJobs, DrmaaError
@@ -23,8 +36,20 @@ class Scheduler(object):
   
   logger = None
 
+  is_sleeping = None
+
   def __init__(self):
     self.parallel_job_submission_info = None
+    self.is_sleeping = False
+
+  def sleep(self):
+    self.is_sleeping = True
+
+  def wake(self):
+    self.is_sleeping = False
+
+  def clean(self):
+    pass
 
   def job_submission(self, job):
     '''
@@ -80,7 +105,10 @@ class Drmaa(Scheduler):
   
   logger = None
 
-  def __init__(self, drmaa_implementation, parallel_job_submission_info):
+  def __init__(self, 
+               drmaa_implementation, 
+               parallel_job_submission_info, 
+               tmp_file_path=None):
 
     self.logger = self.logger = logging.getLogger('ljp.drmaajs')
 
@@ -99,20 +127,50 @@ class Drmaa(Scheduler):
 
     # patch for the PBS-torque DRMAA implementation
     if self._drmaa_implementation == "PBS":
+      if tmp_file_path == None:
+        self.tmp_file_path = os.path.abspath("/tmp")
+      else:
+        self.tmp_file_path = os.path.abspath(tmp_file_path)
       try:
         jobTemplateId = self._drmaa.allocateJobTemplate()
         self._drmaa.setCommand(jobTemplateId, "echo", [])
         self._drmaa.setAttribute(jobTemplateId, 
                                 "drmaa_output_path", 
-                                "[void]:/tmp/soma-workflow-empty-job.o")
+                                "[void]:" + os.path.join(self.tmp_file_path, "soma-workflow-empty-job-patch-torque.o"))
         self._drmaa.setAttribute(jobTemplateId, 
                                 "drmaa_error_path", 
-                                "[void]:/tmp/soma-workflow-empty-job.e")
+                                "[void]:" + os.path.join(self.tmp_file_path, "soma-workflow-empty-job-patch-torque.e"))
         self._drmaa.runJob(jobTemplateId)
       except DrmaaError, e:
         raise DRMError("%s" %e)
       ################################
+
+  def clean(self):
+    if self._drmaa_implementation == "PBS":
+      tmp_out = os.path.join(self.tmp_file_path, "soma-workflow-empty-job-patch-torque.o")
+      tmp_err = os.path.join(self.tmp_file_path, "soma-workflow-empty-job-patch-torque.e")
+      if os.path.isfile(tmp_out):
+        os.remove(tmp_out)
+      if os.path.isfile(tmp_err):
+        os.remove(tmp_err)
+        
+  def sleep(self):
+    '''
+    Some Drmaa sessions expire if they idle too long.  
+    '''
+    self.is_sleeping = True
+    self._drmaa = None
     
+  def wake(self):
+    '''
+    Creates a fresh Drmaa session.
+    '''
+    self.is_sleeping = False
+    self._drmaa = DrmaaJobs()
+    try:
+      self._drmaa.initSession()
+    except DrmaaError, e:
+      raise DRMError("Could not create the DRMAA session: %s" %e )
 
   def job_submission(self, job):
     '''
@@ -122,6 +180,8 @@ class Drmaa(Scheduler):
     @return: drmaa job id 
     '''
 
+    if self.is_sleeping: self.wake()
+
     # patch for the PBS-torque DRMAA implementation
     command = []
     job_command = job.plain_command()
@@ -194,6 +254,7 @@ class Drmaa(Scheduler):
     
 
   def get_job_status(self, scheduler_job_id):
+    if self.is_sleeping: self.wake()
     try:
       status = self._drmaa.jobStatus(scheduler_job_id)
     except DrmaaError, e:
@@ -202,6 +263,7 @@ class Drmaa(Scheduler):
 
 
   def get_job_exit_info(self, scheduler_job_id):
+    if self.is_sleeping: self.wake()
     exit_status, exit_value, term_sig, resource_usage = self._drmaa.wait(scheduler_job_id, 0)
 
     str_rusage = ''
@@ -228,7 +290,7 @@ class Drmaa(Scheduler):
     max_node_num: maximum node number the job requests (on a unique machine or 
     separated machine depending on the parallel configuration)
     ''' 
-
+    if self.is_sleeping: self.wake()
     self.logger.debug(">> _setDrmaaParallelJob")
   
     cluster_specific_cfg_name = self.parallel_job_submission_info[configuration_name]
@@ -258,20 +320,19 @@ class Drmaa(Scheduler):
 
 
   def kill_job(self, scheduler_job_id):
+    if self.is_sleeping: self.wake()
     try:
       self._drmaa.terminate(scheduler_job_id)
     except DrmaaError, e:
       raise DRMError("%s" %e)
+  
 
-
-
-
-class LocalScheduler(object):
+class LocalScheduler(Scheduler):
   '''
   Allow to submit, kill and get the status of jobs.
   Run on one machine without dependencies.
 
-  * _nb_proc *int*
+  * _proc_nb *int*
 
   * _queue *list of scheduler jobs ids*
   
@@ -285,7 +346,7 @@ class LocalScheduler(object):
 
   * _loop *thread*
 
-  * _period *int*
+  * _interval *int*
 
   * _look *threading.RLock*
   '''
@@ -293,7 +354,7 @@ class LocalScheduler(object):
   
   logger = None
 
-  _nb_proc = None
+  _proc_nb = None
 
   _queue = None
 
@@ -307,18 +368,17 @@ class LocalScheduler(object):
 
   _loop = None
 
-  _period = None
+  _interval = None
 
   _lock = None
 
-  def __init__(self, nb_proc=1, period=1):
+  def __init__(self, proc_nb=1, interval=1):
     super(LocalScheduler, self).__init__()
   
-
     self.parallel_job_submission_info = None
 
-    self._nb_proc = nb_proc
-    self._period = period
+    self._proc_nb = proc_nb
+    self._interval = interval
     self._queue = []
     self._jobs = {}
     self._processes = {}
@@ -327,11 +387,13 @@ class LocalScheduler(object):
 
     self._lock = threading.RLock()
 
+    self.stop_thread_loop = False
+
     def loop(self):
-      while True:
+      while not self.stop_thread_loop:
         with self._lock:
           self._iterate()
-        time.sleep(self._period)
+        time.sleep(self._interval)
 
     self._loop = threading.Thread(name="scheduler_loop",
                                   target=loop,
@@ -339,12 +401,26 @@ class LocalScheduler(object):
     self._loop.setDaemon(True)
     self._loop.start()
 
+    atexit.register(LocalScheduler.end_scheduler_thread, self)
+
+  def change_proc_nb(self, proc_nb):
+    with self._lock:
+      self._proc_nb = proc_nb
+
+  def change_interval(self, interval):
+    with self._lock:
+      self._interval = interval
+
+  def end_scheduler_thread(self):
+    with self._lock:
+      self.stop_thread_loop = True
+      self._loop.join()
+      print "Soma scheduler thread ended nicely."
 
   def _iterate(self):
     # Nothing to do if the queue is empty and nothing is running
     if not self._queue and not self._processes:
       return
-
     #print "#############################"
     # Control the running jobs
     ended_jobs = []
@@ -366,7 +442,7 @@ class LocalScheduler(object):
 
     # run new jobs
     while self._queue and \
-          len(self._processes) < self._nb_proc:
+          len(self._processes) < self._proc_nb:
       job_id = self._queue.pop(0)
       job = self._jobs[job_id]
       #print "new job " + repr(job.job_id)
@@ -458,7 +534,8 @@ class LocalScheduler(object):
       self._queue.append(job.job_id)
       self._jobs[job.job_id] = job
       self._status[job.job_id] = constants.QUEUED_ACTIVE
-
+      self._queue.sort(key=lambda job_id: self._jobs[job_id].priority, 
+                       reverse=True)
     return job.job_id
 
 
@@ -500,7 +577,21 @@ class LocalScheduler(object):
       #print "kill job " + repr(scheduler_job_id)
       if scheduler_job_id in self._processes:
         #print "    => kill the process "
-        self._processes[scheduler_job_id].kill()
+        process = self._processes[scheduler_job_id]
+        if sys.version_info < (2, 6):
+          if sys.platform == 'win32':
+            PROCESS_TERMINATE = 1
+            handle = ctypes.windll.kernel32.OpenProcess(PROCESS_TERMINATE, 
+                                               False, 
+                                               process.pid)
+            ctypes.windll.kernel32.TerminateProcess(handle, -1)
+            ctypes.windll.kernel32.CloseHandle(handle)
+          else:
+            os.kill(process.pid, signal.SIGKILL)
+            os.wait()
+        else:
+          process.kill()
+
         del self._processes[scheduler_job_id]
         self._status[scheduler_job_id] = constants.FAILED
         self._exit_info[scheduler_job_id] = (constants.USER_KILLED,
@@ -516,3 +607,33 @@ class LocalScheduler(object):
                                               None,
                                               None,
                                               None)
+
+
+class ConfiguredLocalScheduler(LocalScheduler):
+  '''
+  Local scheduler synchronized with a configuration object.
+  '''
+
+  _config = None
+
+  def __init__(self, config):
+    '''
+    * config *LocalSchedulerCfg*
+    '''
+    super(ConfiguredLocalScheduler, self).__init__(config.get_proc_nb(), 
+                                           config.get_interval())
+    self._config = config
+
+    self._config.addObserver(self,
+                             "update_from_config", 
+                             [LocalSchedulerCfg.PROC_NB_CHANGED, LocalSchedulerCfg.INTERVAL_CHANGED])
+
+
+  def update_from_config(self, observable, event, msg):
+    if event == LocalSchedulerCfg.PROC_NB_CHANGED:
+      self.change_proc_nb(self._config.get_proc_nb())
+    if event == LocalSchedulerCfg.PROC_NB_CHANGED:
+      self.change_interval(self._config.get_interval())
+    self._config.save_to_file()
+    
+
